import torchvision.transforms as transforms
import torch
import torch.nn.functional as F


import warnings
warnings.filterwarnings("ignore")

transform_to_tensor = transforms.Compose([
                transforms.ToTensor()
            ])

transform_and_rotate = transforms.Compose([
                transforms.ToPILImage(),
                transforms.RandomRotation(degrees = 5, resample=False, expand=False, center=None, fill=None),
                transforms.ToTensor()
            ])


def resize_4d_tensor_by_factor(x, height_factor, width_factor):
    res = F.interpolate(x, scale_factor= (height_factor, width_factor), mode = 'bilinear')
    return res

def resize_4d_tensor_by_size(x, height, width):
    res = F.interpolate(x, size =  (height, width), mode = 'bilinear')
    return res


import torch.nn as nn
import torch
import numpy as np

from .utils import init_image_param
import torchvision.transforms as transforms

from .utils import fft_to_rgb, lucid_colorspace_to_rgb, normalize

class BaseImageParam(nn.Module):
    def __init__(self):
        super().__init__()
        self.height = None
        self.width = None
        self.param = None
        self.optimizer = None
    
    def forward(self):
        """This is what the model gets, should be processed and normalized with the right values

        The model gets:  self.normalize(self.postprocess(self.param))

        Raises:
            NotImplementedError: Implemented below, you're in the base class.
        """
        raise NotImplementedError

    def postprocess(self):
        """Moves the image from the frequency domain to Spatial (Visible to the eyes)

        Raises:
            NotImplementedError: Implemented below, you're in the base class.
        """
        raise NotImplementedError

    def normalize(self):
        """Normalizing wrapper, you can either use torchvision.transforms.Normalize() or something else

        Raises:
            NotImplementedError: Implemented below, you're in the base class.
        """
        raise NotImplementedError
        
    def fetch_optimizer(self, params_list, optimizer = None, lr = 1e-3, weight_decay = 0.):
        if optimizer is not None:
            optimizer = optimizer(params_list, lr = lr, weight_decay = weight_decay)
        else:
            optimizer = torch.optim.AdamW(params_list, lr=lr, weight_decay=weight_decay)
        return optimizer

    def get_optimizer(self, lr, weight_decay):

        self.optimizer = self.fetch_optimizer(
                                    params_list = [self.param],
                                    lr = lr,
                                    weight_decay= weight_decay
                                    )
    
    def clip_grads(self, grad_clip = 1.):
        torch.nn.utils.clip_grad_norm_(self.param,grad_clip)

    def to_hwc_tensor(self, device = 'cpu'):
        rgb = self.postprocess(device = device)[0].permute(1,2,0).detach()
        return rgb

    def to_chw_tensor(self, device = 'cpu'):
        t = self.postprocess(device = device)[0].detach()
        return t

    def __array__(self):
        """Generally used for plt.imshow(), converts the image parameter to a numpy array

        Returns:
            numpy.ndarray
        """
        return self.to_hwc_tensor().numpy()

    def save(self, filename):
        """Save an image_param as an image. Uses PIL to save the image

        usage:
            
            image_param.save(filename = 'my_image.jpg')
            
        Args:
            filename (str): image.jpg
        """
        ten = self.to_chw_tensor()
        im = transforms.ToPILImage()(ten)
        im.save(filename)


class auto_image_param(BaseImageParam):
    """Trainable image parameter which can be used to activate 
           different parts of a neural net

        Args:
            height (int): Height of image
            width (int): Width of image
            device (str): 'cpu' or 'cuda'
            standard_deviation (float): Standard deviation of the image initiated in the frequency domain. ). 0.01 works well
    """
    def __init__(self, height, width, device, standard_deviation):
        
        super().__init__()
        self.height = height
        self.width = width

        '''
        odd width is resized to even with one extra column
        '''
        if self.width %2 ==1:
            self.param = init_image_param(height = self.height, width = width + 1, sd = standard_deviation, device = device)
        else:
            self.param = init_image_param(height = self.height, width = self.width, sd = standard_deviation, device = device)

        self.param.requires_grad_()
        self.optimizer = None

    def postprocess(self, device):
        img = fft_to_rgb(height = self.height, width = self.width,  image_parameter = self.param, device= device)
        img = lucid_colorspace_to_rgb(t = img, device = device)
        img = torch.sigmoid(img)
        return img

    def normalize(self,x, device):
        return normalize(x = x, device= device)

    def forward(self, device):
        return self.normalize(self.postprocess(device = device), device= device)



import numpy as np
import torch
from torch import tensor
from torchvision import transforms

from .image_transforms import resize_4d_tensor_by_size
from .error_handlers import PytorchVersionError
from .constants import Constants


def init_image_param(height , width, sd=0.01, device = 'cuda'):
    """Initializes an image parameter in the frequency domain

    Args:
        height (int): height of image
        width (int): width of image
        sd (float, optional): Standard deviation of pixel values. Defaults to 0.01.
        device (str): 'cpu' or 'cuda'

    Returns:
        torch.tensor: image param to backpropagate on
    """
    img_buf = np.random.normal(size=(1, 3, height, width), scale=sd).astype(np.float32)
    spectrum_t = tensor(img_buf).float().to(device)
    return spectrum_t

def get_fft_scale(h, w, decay_power=.75, device = 'cuda'):
    d=.5**.5 # set center frequency scale to 1
    fy = np.fft.fftfreq(h,d=d)[:,None]

    if w %2 ==1:
        fx = np.fft.rfftfreq(w,d=d)[: (w+1) // 2]   
    else:

        fx = np.fft.rfftfreq(w,d=d)[: w // 2]     

    freqs = (fx*fx + fy*fy) ** decay_power
    scale = 1.0 / np.maximum(freqs, 1.0 / (max(w, h)*d))
    scale = tensor(scale).float().to(device)

    return scale


def fft_to_rgb(height, width, image_parameter, device = 'cuda'):
    """convert image param to NCHW 

    WARNING: torch v1.7.0 works differently from torch v1.8.0 on fft. 
    torch-dreams supports ONLY 1.8.x 

    Latest docs: https://pytorch.org/docs/stable/fft.html

    Also refer:
        https://github.com/pytorch/pytorch/issues/49637

    Args:
        height (int): height of image
        width (int): width of image 
        image_parameter (auto_image_param): auto_image_param.param

    Returns:
        torch.tensor: NCHW tensor

    """
    scale = get_fft_scale(height, width, device= device).to(image_parameter.device)
    # print(scale.shape, image_parameter.shape)
    if width %2 ==1:
        image_parameter = image_parameter.reshape(1,3,height, (width+1)//2, 2)
    else:
        image_parameter = image_parameter.reshape(1,3,height, width//2, 2)

    image_parameter = torch.complex(image_parameter[..., 0], image_parameter[..., 1])
    t = scale * image_parameter
   
    if  torch.__version__[:3] == '1.8':

        t = torch.fft.irfft2(t,  s = (height, width), norm = 'ortho')
    else:
        raise PytorchVersionError(version = torch.__version__)

    return t


def lucid_colorspace_to_rgb(t,device = 'cuda'):

    t_flat = t.permute(0,2,3,1)
    t_flat = torch.matmul(t_flat.to(device) , Constants.color_correlation_matrix.T.to(device))
    t = t_flat.permute(0,3,1,2)
    return t


def normalize(x, device = 'cuda'):
    return (x-Constants.imagenet_mean[...,None,None].to(device)) / Constants.imagenet_std[...,None,None].to(device)

def get_fft_scale_custom_img(h, w, decay_power=.75, device = 'cuda'):
    d=.5**.5 # set center frequency scale to 1
    fy = np.fft.fftfreq(h,d=d)[:,None]

    fx = np.fft.rfftfreq(w,d=d)[:(w//2)+1]   
    freqs = (fx*fx + fy*fy) ** decay_power
    scale = 1.0 / np.maximum(freqs, 1.0 / (max(w, h)*d))
    scale = torch.tensor(scale).float().to(device)

    return scale

def denormalize(x):

    return x.float()*Constants.imagenet_std[...,None,None].to(x.device) + Constants.imagenet_mean[...,None,None].to(x.device)

def rgb_to_lucid_colorspace(t, device = 'cuda'):
    t_flat = t.permute(0,2,3,1)
    inverse = torch.inverse(Constants.color_correlation_matrix.T.to(device))
    t_flat = torch.matmul(t_flat.to(device), inverse)
    t = t_flat.permute(0,3,1,2)
    return t

def chw_rgb_to_fft_param(x, device):
    im_tensor = torch.tensor(x).unsqueeze(0).float()

    x = rgb_to_lucid_colorspace(denormalize(im_tensor), device= device)

    x = torch.fft.rfft2(x, s = (x.shape[-2], x.shape[-1]), norm = 'ortho')
    return x

def fft_to_rgb_custom_img(height, width, image_parameter, device = 'cuda'):

    scale = get_fft_scale_custom_img(height, width , device= device).to(image_parameter.device)
    t = scale * image_parameter
   
    if  torch.__version__[:3] == '1.8':
        t = torch.fft.irfft2(t,  s = (height, width), norm = 'ortho')
    else:
        raise PytorchVersionError(version = torch.__version__)

    return t    


class PytorchVersionError(Exception):
    """Raised when the user is not on pytorch 1.8.x

    Args:
        version (str): torch.__version__
    """ 
    def __init__(self, version):
        self.version = version
        self.message = "Expected pytorch to have version 1.8.x but got: " + self.version + "\n Please consider updating pytorch from: https://pytorch.org/get-started/locally"
    def __str__(self):
        return self.message


import cv2
import torch
from .custom_image_param import custom_image_param
from .transforms import imagenet_transform

class masked_image_param(custom_image_param):
    def __init__(self, mask_tensor, image = None , device = 'cuda'):
        """Custom image param, but with a mask over the original image. 

        The mask helps update only certain parts of the image 
        and leave the rest untouched by training. Ideal for 
        things like: "differentiable" backgrounds.

        Args:
            image (str or torch.tensor): "path/to/image.jpg" or NCHW tensor
            mask_tensor (torch.tensor): NCHW tensor whose values are clipped between 0,1
            device (str): 'cpu' or 'cuda'
        """

        self.width = mask_tensor.shape[-1]
        self.height = mask_tensor.shape[-2]

        if image is None:
            from .utils import init_image_param, fft_to_rgb, lucid_colorspace_to_rgb
            if self.width %2 ==1:
                self.param = init_image_param(height = self.height, width = self.width + 1, sd = 0.01, device = device)
            else:
                self.param = init_image_param(height = self.height, width = self.width, sd = 0.01, device = device)
            img = fft_to_rgb(height = self.height, width = self.width,  image_parameter = self.param, device= device)
            img = lucid_colorspace_to_rgb(t = img, device = device)
            image = torch.sigmoid(img)


            # image = torch.ones(1,3,self.height,self.width)

        super().__init__(image = image, device= device)

        self.mask = mask_tensor.to(self.device)

        if isinstance(image, str):
            image = cv2.cvtColor(cv2.imread(image), cv2.COLOR_BGR2RGB)/255.
            image = torch.tensor(image).permute(-1,0,1).unsqueeze(0)

        self.original_nchw_image_tensor = image.to(device)

        assert self.mask.shape[-2:] == self.to_nchw_tensor(device = self.device).shape[-2:], "The height and width of the input image" + str(self.to_nchw_tensor(device = self.device).shape[-2:]) + "and the mask" + str(self.mask.shape[-2:]) + "do not match."

    def to_chw_tensor(self, device = 'cpu'):
        t = self.forward(device= device).squeeze(0).clamp(0,1).detach()  + self.original_nchw_image_tensor.to(device) * (1-self.mask.to(device)) 
        return t.squeeze(0)

    def forward(self, device):
        return self.normalize(self.postprocess(device = device), device= device).clamp(0,1) * self.mask.to(device) 

    def update_mask(self, mask):
        """updates the mask to have a new value

        Warning: it also updates the original image 

        Args:
            mask (torch.tensor): new mask to be used
        """

        self.original_nchw_image_tensor = self.to_chw_tensor(device = self.device).unsqueeze(0)
        self.mask = mask.to(self.device)





import torchvision.transforms as transforms
import torch.nn as nn
import random
from .image_transforms import resize_4d_tensor_by_factor, resize_4d_tensor_by_size

imagenet_transform = transforms.Normalize(
    mean=[0.485, 0.456, 0.406], 
    std= [0.229, 0.224, 0.225]
)


class random_resize(nn.Module):
    def __init__(self, max_size_factor, min_size_factor):
        super().__init__()
        self.max_size_factor = max_size_factor
        self.min_size_factor = min_size_factor

    def forward(self, x):
    
        # size = random.randint(a = 300, b = 600)
        # resized= resize_4d_tensor_by_size(x = x, height = size, width = size)

        height_factor = random.uniform(a = self.min_size_factor , b = self.max_size_factor)
        width_factor = random.uniform(a = self.min_size_factor , b = self.max_size_factor)

        resized = resize_4d_tensor_by_factor(x = x, height_factor = height_factor, width_factor = width_factor)
        return resized

class pair_random_resize(nn.Module):
    def __init__(self, max_size_factor, min_size_factor):
        super().__init__()
        self.max_size_factor = max_size_factor
        self.min_size_factor = min_size_factor

    def forward(self, tensors = []):
    
        height_factor = random.uniform(a = self.min_size_factor , b = self.max_size_factor)
        width_factor = random.uniform(a = self.min_size_factor , b = self.max_size_factor)

        outputs = []
        for x in tensors:
            resized_tensor = resize_4d_tensor_by_factor(x = x, height_factor = height_factor, width_factor = width_factor)
            outputs.append(resized_tensor)

        return outputs

class pair_random_affine(nn.Module):
    def __init__(self, degrees, translate_x, translate_y):
        super().__init__()
        self.degrees = degrees
        self.translate_x = translate_x
        self.translate_y = translate_y
        self.affine = transforms.RandomAffine(degrees = self.degrees, translate= (self.translate_x, self.translate_y))

    def forward(self, tensors = []):

        params = self.affine.get_params(degrees = (-self.degrees, self.degrees),  translate= (self.translate_x, self.translate_y), scale_ranges = (1,1), shears = (0,0), img_size = (tensors[0].shape[-2], tensors[0].shape[1]))

        outputs = []
        for x in tensors:
            affined = transforms.functional.affine(x, *params)
            outputs.append(affined)

        return outputs
                


import torch.nn as nn
import torch

from .image_transforms import resize_4d_tensor_by_size

class CaricatureLoss(nn.Module):

    def __init__(self, power = 1.):
        super().__init__()        
        self.power = power 

    def cosine_dissimilarity(self,x,y, eps = 1e-6):
        """
        tried my best to replicate: 
            https://github.com/tensorflow/lucid/blob/6dcc927e4ff4e7ef4d9c54d27b0352849dadd1bb/lucid/recipes/caricature.py#L21
            
        if I missed something out, please get in touch with me on Distill slack: @Mayukh 

        or email me: 
            mayukhmainak2000@gmail.com
        or find me on github: 
            github.com/mayukhdeb
        """

        if x.shape != y.shape:
            """
            if their shapes are not equal (likely due to using static caricatures), then resize the target accordingly
            """
            y = resize_4d_tensor_by_size(y.unsqueeze(0), height = x.shape[-2], width = x.shape[-1] ).squeeze(0)

        y = y.detach()
        
        numerator = (x*y.detach()).sum() 
        denominator = torch.sqrt((y**2).sum()) + eps
        cossim = numerator/denominator
        cossim = torch.maximum(torch.tensor(0.1).to(cossim.device), cossim)
        loss = -(cossim*numerator**self.power)
        return loss

    def loss(self, x,y):
        loss = self.cosine_dissimilarity(x,y)  
        return loss

    def forward(self,layer_outputs, ideal_layer_outputs):
        assert len(layer_outputs) == len(ideal_layer_outputs)
        loss = 0.
        for i  in range(len(layer_outputs)):
        
            l  = self.loss(layer_outputs[i] ,ideal_layer_outputs[i])
            loss += l
           
        # print(loss)
        return loss 